/*
 * Octopus / Cadmus2 USB Adaptation Layer for Linux
 * Written by Marc Boucher <marc@linuxant.com>
 *
 * Copyright (c) 2002-2003 Linuxant inc.
 * All rights reserved.
 * 
 * TODO:
 * 	- Proper locking
 * 	- DCP support
 * 	- Cleanup
 */


#include "oscompat.h"
#include "ostypedefs.h"
#include "osservices.h"

#include <linux/usb.h>
#include <linux/slab.h>
#include <asm/io.h>
#include <asm/pgtable.h>

#ifdef CONFIG_ALL_PPC
#include <asm/machdep.h>
#include <asm/pmac_feature.h>
#endif

#include "usbhalos.h"
#include "c2firmware.h"

#define	MAX_NOTIFY_DATA_PACKET	64
#define	MAX_BULK_DATA_PACKET	64
#define	MAX_CONTROL_DATA_PACKET	64
#define DATA_PIPE_NUM			4
#define INT_PIPE_NUM			1

#define MAX_RECEIVE_URB_BUFSIZE	(MAX_BULK_DATA_PACKET*4)
#define MAX_SEND_URB_BUFSIZE	(MAX_BULK_DATA_PACKET*4)
#define MAX_NOTIFY_URB_BUFSIZE	MAX_NOTIFY_PACKET
#define MAX_CONTROL_URB_BUFSIZE	(MAX_BULK_DATA_PACKET*4)
#define MAX_DOWNLOAD_URB_BUFSIZE (8*1024)
#define MAX_DCP_URB_BUFSIZE		MAX_BULK_DATA_PACKET

#define NUM_RECEIVE_URBS		2
#define NUM_SEND_URBS			2
#define NUM_NOTIFY_URBS			1
#define NUM_CONTROL_URBS		2
#define NUM_DOWNLOAD_URBS		2

#ifdef _DEBUG
#define SIGNATURE 'C2OS'
#define CHECK_SIGNATURE(instance) { ASSERT(instance); ASSERT (instance->uSignature == SIGNATURE); }
#define BREAKPOINT(msg)     { err(msg); KDB_ENTER();   }
#define ASSERT(x) { if(!(x)) { err("Assertion (%s) failed in %s() (%s:%d)", #x, __FUNCTION__, __FILE__, __LINE__); BREAKPOINT("Assert"); } }
#else
#define CHECK_SIGNATURE(instance)
#define BREAKPOINT(msg)
#define ASSERT(x)
#endif

//#undef dbg
//#define dbg(format, arg...) { UINT32 dwTime = OsGetSystemTime(); printk(KERN_DEBUG "%07d.%03d: "__FILE__ ": " format "\n" , dwTime/1000, dwTime%1000, ## arg); }
//#define dbg(format, arg...) do {} while (0)

#define Working TRUE
#define GET_DEVICE_PNP_STATE(_USBOSHAL_) ((_USBOSHAL_) && usb_started)

#define IsKMallocMem(vaddr) (((void *)vaddr) < high_memory)

typedef struct urb *PURB;

typedef struct {
	wait_queue_head_t wq;
	int flag;
	spinlock_t lock;
} EVENT, *PEVENT;

typedef enum {
	ULE_FREE,
	ULE_ALLOCATED,
	ULE_QUEUED,
	ULE_SUBMITTED,
	ULE_COMPLETED,
	ULE_ABORTED,
	ULE_FAILED,
	ULE_LAST
} URB_LIST_ENTRY_STATE;

static const char *uleStateNames[] = {
	"FREE",
	"ALLOCATED",
	"PENDING",
	"SUBMITTED",
	"COMPLETED",
	"ABORTED",
	"FAILED",
	"LAST"
};

struct tagUSBOSHAL;
typedef struct {
	struct list_head entry;
	struct list_head queued_entry;
	struct tagURB_LIST *ul;
	struct tagUSBOSHAL *pUsbOsHal;
	URB_LIST_ENTRY_STATE state;
	PURB volatile urb;
	struct usb_ctrlrequest	devRequest;
	PUINT8 bufferPtr;
	size_t bufferSize;
	PUINT8 userBufferPtr;
	size_t userBufferSize;
	UINT32	startTime;
} URB_LIST_ENTRY, *PURB_LIST_ENTRY;

typedef struct tagURB_LIST {
	struct list_head urb_list;
	struct list_head queued_list;
	volatile int count[ULE_LAST];
} URB_LIST, *PURB_LIST;

//
// A structure representing the instance information associated with
// this particular device.
//
typedef struct tagUSBOSHAL {
	struct list_head instance_list_entry;

	BOOL (*cbUsbHalNotificationReceived)	(PVOID hUpper, PUINT8 pDataReceived, unsigned RecvLength);
	void (*cbUsbHalDataReceived)			(PVOID hUpper, PUINT8 pDataReceived, unsigned RecvLength);
	void (*cbUsbHalDataTransmitDone)		(PVOID hUpper);
	void (*cbUsbHalControlTransferDone)		(PVOID hUpper, UINT8 Request, UINT16 value, UINT16 index);
	void (*cbUsbHalDownloadTransmitDone)	(PVOID hUpper);


	EVENT				ControlRequestEvent;

	struct usb_device	*pUsbDevice;
	struct usb_interface	*pUsbDataInterface;
	struct usb_interface	*pUsbCommInterface;

	// URBs for 
	// data transmit pipe
	URB_LIST SendUrbList;
	// data receive pipe
	URB_LIST RecvUrbList;
	// notify pipe
	URB_LIST NotifyUrbList;
	// control pipe
	URB_LIST ControlUrbList;
	// datapump download
	URB_LIST DownloadUrbList;
#ifdef USE_DCP
	// datapump download
	URB_LIST DcpUrbList;
#endif

	int				DataInPipe,
					DataOutPipe,
					DcpInPipe,
					DownloadPipe,
					NotifyPipe;

	__u8			NotifyInterval;

	HANDLE hUpper; // handle of OS independent portion of the USB HAL

	BOOL			UpdateEEPROM;

//	BOOL				bPipeResetInProgress;
//    int		        ResetAttemptCount;

	PVOID				pLastControlData;
	size_t				nLastControlSize;

	size_t				MaxControlSize;
	size_t				MaxNotifySize;

#ifdef _DEBUG
	ULONG		uSignature;		
#endif
} USBOSHAL, *PUSBOSHAL;

/* table of devices that work with this driver */
static struct usb_device_id usbhcf_table [] = {
	{ USB_DEVICE(USB_AAPL_VENDOR_ID, USB_AAPL_PRODUCT_ID) },
	{ USB_DEVICE(USB_CNXT_VENDOR_ID, USB_CNXT_PRODUCT_ID) },
	{ }
};

MODULE_DEVICE_TABLE (usb, usbhcf_table);

static void usbhcf_disconnect(struct usb_device *dev, void *ptr);
static void *usbhcf_probe(struct usb_device *dev, unsigned int ifnum, const struct usb_device_id *id);


static struct usb_driver usbhcf_driver = {
	name:		"usbhcf",
	probe:		usbhcf_probe,
	disconnect:	usbhcf_disconnect,
	id_table:	usbhcf_table,
};


static BOOL usb_started;

static LIST_HEAD(instance_list);

#if 0
// 20 times pipe can be stalled and reset
#define ALLOWED_PIPE_RESET_ATTEMPTS 20
#endif

static void	SubmitNextUrbListEntry(PURB_LIST_ENTRY ule);

static inline void
ChangeUrbListEntryState(PURB_LIST_ENTRY ule, URB_LIST_ENTRY_STATE newstate)
{
	URB_LIST_ENTRY_STATE oldstate;

	dbg("%s: ule=%p %s -> %s", __FUNCTION__, ule,
			uleStateNames[ule->state],
		   	uleStateNames[newstate]);
#if 0
	{
		int i;

		for(i = 0; i < ULE_LAST; i++) {
			dbg("%s: ul=%p %s=%d", __FUNCTION__, ule->ul,
					uleStateNames[i], ule->ul->count[i]);
		}
	}
#endif

	oldstate = ule->state;
	if(oldstate == newstate)
		return;

	ule->ul->count[ule->state]--;
	ule->state = newstate;
	ule->ul->count[ule->state]++;


	if(oldstate == ULE_QUEUED) {
		list_del(&ule->queued_entry);
	}

	if(newstate == ULE_SUBMITTED) {
		MOD_INC_USE_COUNT;
//warn("%s: aft inc use_count=%d", __FUNCTION__, GET_USE_COUNT(THIS_MODULE));
	} else if(oldstate == ULE_SUBMITTED) {
		MOD_DEC_USE_COUNT;
//warn("%s: aft dec use_count=%d", __FUNCTION__, GET_USE_COUNT(THIS_MODULE));
		if((newstate == ULE_COMPLETED) || (newstate == ULE_FAILED))
			SubmitNextUrbListEntry(ule);
	}

	if(newstate >= ULE_COMPLETED) {
		ChangeUrbListEntryState(ule, ULE_FREE);
	}
}

static BOOL
SubmitUrbListEntry(PURB_LIST_ENTRY ule)
{
	int res;

	if(0 == ule->ul->count[ULE_SUBMITTED]) {
		ule->startTime = OsGetSystemTime();
		ChangeUrbListEntryState(ule, ULE_SUBMITTED);
		if ((res=usb_submit_urb(ule->urb))) {
			warn("%s: submit failed: %d", __FUNCTION__, res);
			ChangeUrbListEntryState(ule, ULE_FREE);
			return FALSE;
		}
	} else {
		ChangeUrbListEntryState(ule, ULE_QUEUED);
		list_add_tail(&ule->queued_entry, &ule->ul->queued_list);
		dbg("%s: submitted count=%d, added %p to pending queue (count=%d)",
				__FUNCTION__,
				ule->ul->count[ULE_SUBMITTED],
				ule,
				ule->ul->count[ULE_QUEUED]);
	}

	return TRUE;
}


static void	SubmitNextUrbListEntry(PURB_LIST_ENTRY ule)
{
	if((0 != ule->ul->count[ULE_QUEUED]) &&
			(0 == ule->ul->count[ULE_SUBMITTED])) {
		PURB_LIST_ENTRY next_ule;

		ASSERT(!list_empty(&ule->ul->queued_list));

		next_ule = list_entry(ule->ul->queued_list.next, URB_LIST_ENTRY, queued_entry);
		dbg("%s: %p", __FUNCTION__, next_ule);

		SubmitUrbListEntry(next_ule);
	}
}

static PURB_LIST_ENTRY
GetFreeUrbFromList(PURB_LIST ul)
{
	struct list_head *e;
	PURB_LIST_ENTRY ule;

	list_for_each (e, &ul->urb_list) {
		ule = list_entry(e, URB_LIST_ENTRY, entry);
		if(ule->state == ULE_FREE) {
			ChangeUrbListEntryState(ule, ULE_ALLOCATED);

			ule->userBufferPtr = NULL;
			ule->userBufferSize = 0;
			ule->startTime = 0;
			dbg("%s: ul=%p ule=%p", __FUNCTION__, ul, ule);
			return ule;
		}

	}
	//warn("%s: no more (ul=%p)", __FUNCTION__, ul);
	return NULL;
}

static PURB_LIST_ENTRY AllocateUrbListEntry(PURB_LIST ul, PUSBOSHAL pUsbOsHal, size_t bufferSize)
{
	PURB_LIST_ENTRY ule;

	ule = (PURB_LIST_ENTRY) kmalloc(sizeof(URB_LIST_ENTRY) + bufferSize, GFP_KERNEL);
	if(!ule) {
		err("%s: kmalloc failed", __FUNCTION__);
		return NULL;
	}

	memset(ule, 0, sizeof(*ule));
	ule->state = ULE_FREE;
	ule->ul = ul;
	ule->pUsbOsHal = pUsbOsHal;
	ule->urb = usb_alloc_urb(0);
	if(!ule->urb) {
		err("%s: usb_alloc_urb failed", __FUNCTION__);
		kfree(ule);
		return NULL;
	}

	if(bufferSize) {
		ule->bufferPtr = (PUINT8)(ule + 1);
		ule->bufferSize = bufferSize;
	}

	dbg("%s: ul=%p ule=%p bufsize=%d", __FUNCTION__, ul, ule, bufferSize);
	return ule;
}

static void FreeUrbListEntry(PURB_LIST_ENTRY ule)
{
	dbg("%s: ule=%p", __FUNCTION__, ule);
	usb_free_urb(ule->urb);
	kfree(ule);
}

static void AbortListUrbs(PURB_LIST ul)
{
	struct list_head *e;
	PURB_LIST_ENTRY ule;

	list_for_each (e, &ul->urb_list) {
		ule = list_entry(e, URB_LIST_ENTRY, entry);
		if(ule->state != ULE_FREE) {
			dbg("%s: ule %p (urb %p) was still in state %s\n", __FUNCTION__,
					ule, ule->urb, uleStateNames[ule->state]);
			if(ule->state == ULE_SUBMITTED) {
				ule->urb->transfer_flags &= ~USB_ASYNC_UNLINK;
				usb_unlink_urb(ule->urb);
			} else
				ChangeUrbListEntryState(ule, ULE_ABORTED);
		}
	}
}

static void FreeUrbList(PURB_LIST ul)
{
	struct list_head *e, *n;
	PURB_LIST_ENTRY ule;

	AbortListUrbs(ul);
	list_for_each_safe (e, n, &ul->urb_list) {
		ule = list_entry(e, URB_LIST_ENTRY, entry);
		if(ule->state != ULE_FREE) {
			warn("%s: ule %p (urb %p) was in state %s\n", __FUNCTION__,
					ule, ule->urb, uleStateNames[ule->state]);
			ule->urb->transfer_flags &= ~USB_ASYNC_UNLINK;
			usb_unlink_urb(ule->urb);
		}
		ul->count[ule->state]--;
		list_del(&ule->entry);
		FreeUrbListEntry(ule);
	}
}

static BOOL InitUrbList(PURB_LIST ul, PUSBOSHAL pUsbOsHal, size_t bufferSize, int nurbs)
{
	INIT_LIST_HEAD(&ul->urb_list);
	INIT_LIST_HEAD(&ul->queued_list);

	dbg("%s: ul=%p bufsize=%d nurbs=%d", __FUNCTION__, ul, bufferSize, nurbs);
	while(nurbs--) {
		PURB_LIST_ENTRY ule = AllocateUrbListEntry(ul, pUsbOsHal, bufferSize);
		if(!ule) {
			FreeUrbList(ul);
			return FALSE;
		}
		list_add(&ule->entry, &ul->urb_list);
		ul->count[ULE_FREE]++;
	}

	return TRUE;
}


/******************************************************************************
 two functions below indicate to lower HAL that the modem instance is being allocated or
 deallocated. They DO NOT cause alocation/deallocation of the lower HAL which
 typically happens when device is plugged in or out
******************************************************************************/


/******************************************************************************

Name:           OsUsbCreateInstance

Description:    Indicate to the bottom HAL that modem instance is being created

Arguments:      pDeviceID	- OS specific unique ID of a device opened, such as PnP ID
				hUpper		- opaque handle to the top HAL for callbacks

Return Value:   New bottom HAL handle, NULL if failure

Notes:

******************************************************************************/
HANDLE OsUsbCreateInstance (PVOID pDeviceID, HANDLE hUpper,
		BOOL (*cbUsbHalNotificationReceived) (PVOID hUpper, PUINT8 pDataReceived, unsigned RecvLength),
		void (*cbUsbHalDataReceived)			(PVOID hUpper, PUINT8 pDataReceived, unsigned RecvLength),
		void (*cbUsbHalDataTransmitDone)		(PVOID hUpper),
		void (*cbUsbHalControlTransferDone)	(PVOID hUpper, UINT8 Request, UINT16 value, UINT16 index),
		void (*cbUsbHalDownloadTransmitDone) (PVOID hUpper),
		void (*cbUsbEEPROM_Restore)	(HANDLE hOsUsb))
{
	PUSBOSHAL			pUsbOsHal = (PUSBOSHAL)pDeviceID;

	dbg("%s: pDeviceID=%p hUpper=%p", __FUNCTION__, pDeviceID, hUpper);
#ifdef _DEBUG
	pUsbOsHal->uSignature = SIGNATURE;
#endif
	pUsbOsHal->hUpper = hUpper;

	pUsbOsHal->cbUsbHalNotificationReceived = cbUsbHalNotificationReceived;
	pUsbOsHal->cbUsbHalDataReceived = cbUsbHalDataReceived;
	pUsbOsHal->cbUsbHalDataTransmitDone = cbUsbHalDataTransmitDone;
	pUsbOsHal->cbUsbHalControlTransferDone = cbUsbHalControlTransferDone;
	pUsbOsHal->cbUsbHalDownloadTransmitDone = cbUsbHalDownloadTransmitDone;

	if(pUsbOsHal->UpdateEEPROM) {
		info("Restoring USBEEPROM");
			cbUsbEEPROM_Restore (pUsbOsHal);
	}

    usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DataInPipe);
    usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DataOutPipe);
	usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DcpInPipe);
	usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DownloadPipe);
    usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->NotifyPipe);
   
	MOD_INC_USE_COUNT;
//warn("%s: aft inc use_count=%d", __FUNCTION__, GET_USE_COUNT(THIS_MODULE));

	return (HANDLE)pUsbOsHal;
}

/******************************************************************************

Name:           OsUsbDestroyInstance

Description:    Indicate to the bottom HAL that modem instance is being destroyed

Arguments:      hUsbOsHal	- bottom HAL handle

Return Value:   None

Notes:          Top HAL handle is invalid after this call. Bottom HAL must not call
				any callback routines

******************************************************************************/
void OsUsbDestroyInstance (HANDLE hUsbOsHal)
{
	PUSBOSHAL		pUsbOsHal = (PUSBOSHAL)hUsbOsHal;

	dbg("%s: %p", __FUNCTION__, pUsbOsHal);

	pUsbOsHal->hUpper = 0;

	if (Working == GET_DEVICE_PNP_STATE(pUsbOsHal))
	{
		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DataInPipe);
		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DataOutPipe);
		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DcpInPipe);
		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DownloadPipe);
		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->NotifyPipe);
		OsSetEvent ((HOSEVENT)&pUsbOsHal->ControlRequestEvent);
	}

	AbortListUrbs(&pUsbOsHal->NotifyUrbList);
	AbortListUrbs(&pUsbOsHal->RecvUrbList);
	AbortListUrbs(&pUsbOsHal->SendUrbList);
	AbortListUrbs(&pUsbOsHal->ControlUrbList);
	AbortListUrbs(&pUsbOsHal->DownloadUrbList);
#ifdef USE_DCP
	AbortListUrbs(&pUsbOsHal->DcpUrbList);
#endif

	// IB: Need to stop everything here
	OsSleep(10);

	MOD_DEC_USE_COUNT;
//warn("%s: aft dec use_count=%d", __FUNCTION__, GET_USE_COUNT(THIS_MODULE));
}

HANDLE OsUsbGetInstance(UINT32 Unit)
{
	struct list_head *list = &instance_list;

	while (((list = list->next) != &instance_list) && Unit--);
   
	return (list == &instance_list) ? NULL : list_entry(list, USBOSHAL, instance_list_entry);
}

//It downloads FW including script engine.
//The FW starts after been loaded.
//Do not need to wait after download.
static int OsUSBFWDownload (PUSBOSHAL pUsbOsHal)
{
	int i;
	int ret = 0;
	unsigned char info[4];
	int fwSize;
	PUINT8 pC2Firmware;

	dbg("%s: %p", __FUNCTION__, pUsbOsHal);

	for (i=0; i<3; i++)			// make 3 attempts
	{
		ret = usb_control_msg(pUsbOsHal->pUsbDevice,
				usb_rcvctrlpipe(pUsbOsHal->pUsbDevice, 0),
				CD2_GET_INFROMATION,
				USB_TYPE_VENDOR | USB_DIR_IN | USB_RECIP_INTERFACE,
				0,
				i,
				info,
				sizeof(info),
				1*HZ);

		if (ret == sizeof(info))
		{
			dbg("GET INFO BEFORE DOWNLOAD: %d",ret);
			ret = 0;
			break;
		}
	}
	if (i == 3)
	{
		err("CD2_GET_INFROMATION returned error %d", ret);
		return ret;
	}

	dbg("FW Information: %x %x %x %x",info[0],info[1],info[2],info[3]);

    if(info[2] == 0x1)
    {
		switch (info[3])
		{
		case 3:		// X4
			pC2Firmware = C2FirmwareX4;
			fwSize = sizeof(C2FirmwareX4);
			dbg("Uploading C2FirmwareX4 size=%d", fwSize);
			break;
		default:
			err("Unrecognized minor version %x", info[3]);
			return -EINVAL;
		}

		i = 0;
		while (i<fwSize) {
			ULONG nBytes = fwSize - i;
			
			if(nBytes > FW_BLOCK_SIZE)
				nBytes = FW_BLOCK_SIZE;

			if (!OsUsbMakeControlRequestSync((HANDLE)pUsbOsHal, 0,
					CD2_UPLOAD_FIRMWARE, fwSize, i, pC2Firmware+i, &nBytes)) {
				err("CD2_UPLOAD_FIRMWARE error nBytes=%ld i=%d", nBytes, i);
				return ret;
			}

			i += nBytes;
        }

        for (i=0; i<10; i++)
        {
			OsSleep(100);

			ret = usb_control_msg(pUsbOsHal->pUsbDevice,
					usb_rcvctrlpipe(pUsbOsHal->pUsbDevice,0),
					CD2_GET_INFROMATION,
					USB_TYPE_VENDOR | USB_DIR_IN | USB_RECIP_INTERFACE,
					0,
					i,
					info,
					sizeof(info),
					1*HZ);

			if (ret == sizeof(info))
			{
				dbg("GET INFO AFTER DOWNLOAD: %x %x", ret, info[2]);
				ret = 0;
				break;
			}

			dbg("GET INFO AFTER DOWNLOAD err=%d", ret);
		}
           
		if(i != 10) {
           if(info[2] != 0x2)    // 0x2 - HCF, 0x3 - HSF
           {
              ret = -EINVAL;
          	  err("ERROR!!! Wrong Firmware is loaded: %x.",info[2]);
           }
		}
    }
    else
    {
		dbg("Firmware is ALREADY loaded.");
		ret = 0;
    }
    
	dbg("-USBHal_FWDownload");
    return ret;
}

/******************************************************************************

Name:           OsUsbIsStarted

Description:    Query if device is configured and is ready to accept USB transactions

Arguments:      hUsbOsHal	- bottom HAL handle

Return Value:   TRUE if device is ready, FALSE if not

Notes:

******************************************************************************/
BOOL OsUsbIsStarted (HANDLE hUsbOsHal)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;

//	dbg("%s: %p returning %s", __FUNCTION__, pUsbOsHal, (Working == GET_DEVICE_PNP_STATE(pUsbOsHal)) ? "TRUE" : "FALSE");

	return (Working == GET_DEVICE_PNP_STATE(pUsbOsHal));
;
}


/******************************************************************************

Name:           OsUsbResetPort

Description:    Reset USB port.

Arguments:      hUsbOsHal	- bottom HAL handle.

Return Value:   None

******************************************************************************/
void OsUsbResetPort(HANDLE hUsbOsHal)
{
#if 1
//	int ret;

//	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;

	dbg("%s: %p", __FUNCTION__, hUsbOsHal);
#if 0
	ret = usb_reset_device(pUsbOsHal->pUsbDevice);
	if(ret) {
		err("OsUsbResetPort: usb_reset_device() failed, ret = %d", ret);
		return;
	}

	ret =  OsUSBFWDownload(pUsbOsHal);
    if (ret)
    {
		err("OsUsbResetPort: OsUSBFWDownload failed, ret = %d", ret);

    }
#endif
#else
	PUSBOSHAL			pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
    NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	pDeviceExtension = pUsbOsHal->pDeviceExtension;
    ULONG				PortStatus = 0;

	ntStatus = CallUSBDI_Sync(pUsbOsHal, IOCTL_INTERNAL_USB_GET_PORT_STATUS, &PortStatus);
	RTRACE2(T__HALOS, 0,"OsUsbResetPort: IOCTL_INTERNAL_USB_GET_PORT_STATUS  NTSTATUS = 0x%x, Port Status = %X", ntStatus, PortStatus);
    if (NT_ERROR(ntStatus))
    {
		RTRACE1(T__HALOS|T__ERROR, 0,"!! OsUsbResetPort:IOCTL_INTERNAL_USB_GET_PORT_STATUS returns error  0x%x", ntStatus);
		return;
    }
	
	if ( !(PortStatus & USBD_PORT_CONNECTED) )
	{
		RTRACE0(T__HALOS|T__ERROR, 0,"!! OsUsbResetPort: Port is NOT connected");
		return;
	}

	if ( PortStatus & USBD_PORT_ENABLED )	// If Port is already enabled, 
	{										// all outstanding requests for a pipe must be canceled before issuing IOCTL_INTERNAL_USB_RESET_PORT.
		RTRACE0(T__HALOS, 0,"OsUsbResetPort: Port is already enabled. ");

		USBMODEM_KillIO (pUsbOsHal);
	}
	ntStatus = CallUSBDI_Sync(pUsbOsHal, IOCTL_INTERNAL_USB_RESET_PORT, NULL);

	RTRACE1(T__HALOS, 0,"OsUsbResetPort: IOCTL_INTERNAL_USB_RESET_PORT  NTSTATUS = 0x%x", ntStatus);

	if (NT_ERROR(ntStatus))
	{
		RTRACE1(T__HALOS|T__ERROR, 0,"!! OsUsbResetPort: IOCTL_INTERNAL_USB_RESET_PORT returns error  0x%x", ntStatus);
		return;
	}

	ntStatus =  OsUSBFWDownload(pUsbOsHal);
    if (NT_ERROR(ntStatus))
    {
		RTRACE1(T__HALOS|T__ERROR, 0,"!! OsUsbResetPort: OsUSBFWDownload returns error  0x%x", ntStatus);
    }
#endif
}


/*****************************************************************************
/
/
/           USB completion routines
/
/
*****************************************************************************/

static void UsbControlCompletionRoutine (PURB urb)
{
	PURB_LIST_ENTRY ule = urb->context;
	PUSBOSHAL pUsbOsHal;
	int UsbStatus = urb->status;
	struct usb_ctrlrequest dr;
    
	dbg("%s: ule=%p urb=%p UsbStatus=%d", __FUNCTION__, ule, urb, UsbStatus);

	if((UsbStatus == -ECONNRESET) || (UsbStatus == -ENOENT)) {
		dbg("%s: -ECONNRESET or -ENOENT", __FUNCTION__);
		ChangeUrbListEntryState(ule, ULE_ABORTED);
		return;
	}

	if(!ule)
		return;

	pUsbOsHal = ule->pUsbOsHal;
	CHECK_SIGNATURE (pUsbOsHal);

	memcpy(&dr, urb->setup_packet, sizeof(dr));

	if (OsGetSystemTime() - ule->startTime > 20)
	{
		warn("%s: control request=%d (index=%d, value=%d) took too long (%ldms)",
				__FUNCTION__,
				dr.bRequest,
				le16_to_cpup(&dr.wIndex),
				le16_to_cpup(&dr.wValue),
				OsGetSystemTime() - ule->startTime);
	}

#if 0
	if (bError && (0x80000004 == urb->UrbHeader.Status) && 
		CD2_CONTROL_SCRIPT == urb->UrbControlVendorClassRequest.Request && 
		3 != urb->UrbControlVendorClassRequest.Index)	// do not restart the transaction if STALL was caused
														// by an attempt to delete script (index = 3)
#endif
#if 0
	// restart if STALLED except if caused by an attempt to
	// delete script (index = 3)
	if ((UsbStatus == -EPIPE) && (CD2_CONTROL_SCRIPT == dr.request) && (3 != le16_to_cpup(&dr.index))) {	
    	urb->next = NULL;
		if(!OsUsbMakeControlRequest (pUsbOsHal, urb, dr.request, le16_to_cpup(&dr.value), le16_to_cpup(&dr.index), pUsbOsHal->pLastControlData, pUsbOsHal->nLastControlSize)) {
			err("%s: Control pipe restart failed", __FUNCTION__);
			BREAKPOINT ("Control pipe restart failed");
		} else {
			// previous call incremented value below
			// we need to put it back
			OsUsbDecrementIoCount(pUsbOsHal);
		}
		return;
	}
#endif

	if (UsbStatus)
	{
//		pUsbOsHal->ResetAttemptCount++;
		err("%s: error: URB status %d, req=%d, index=%d, value=%d",
				__FUNCTION__,
				UsbStatus,
				dr.bRequest,
				le16_to_cpup(&dr.wIndex),
				le16_to_cpup(&dr.wValue));
	}

	OsSetEvent ((HOSEVENT)&pUsbOsHal->ControlRequestEvent);
	if (pUsbOsHal->hUpper)
		pUsbOsHal->cbUsbHalControlTransferDone (pUsbOsHal->hUpper,
				dr.bRequest,
				le16_to_cpup(&dr.wValue),
				le16_to_cpup(&dr.wIndex));

	ChangeUrbListEntryState(ule, UsbStatus ? ULE_FAILED: ULE_COMPLETED);
}

/****************************************************************
 *                                                              *
 *      NAME:                                                   *
 *      PARAMETERS:
 *                                                              *
 *      USAGE:                                                  *
 *                                                              *
 *      DESCRIPTION:                                            *
 *                                                              *
 *      NOTES:          None.                                   *
 ****************************************************************/
static void UsbNotifyCompletionRoutine (PURB urb)
{
	PURB_LIST_ENTRY ule = urb->context;
	PUSBOSHAL pUsbOsHal;
	unsigned char pDataReceived[MAX_NOTIFY_DATA_PACKET];
	int UsbStatus = urb->status, UsbRecvLength = urb->actual_length;
	struct usb_ctrlrequest *dr;
    
	dbg("%s: ule=%p urb=%p UsbStatus=%d UsbRecvLength=%d", __FUNCTION__, ule, urb, UsbStatus, UsbRecvLength);

	if((UsbStatus == -ECONNRESET) || (UsbStatus == -ENOENT)) {
		dbg("%s: -ECONNRESET or -ENOENT", __FUNCTION__);
		ChangeUrbListEntryState(ule, ULE_ABORTED);
		return;
	}

	if(!ule)
		return;

	pUsbOsHal = ule->pUsbOsHal;
	CHECK_SIGNATURE (pUsbOsHal);

	if (UsbStatus)
	{
		err("%s: error: URB status %d", __FUNCTION__, UsbStatus);
		// game over
		usb_unlink_urb(urb);
		return;
	}
	// save a copy of notification data and received length
	ASSERT(UsbRecvLength <= MAX_NOTIFY_DATA_PACKET);
	memcpy (pDataReceived, urb->transfer_buffer, UsbRecvLength); 

	dr = (struct usb_ctrlrequest *)pDataReceived;
	dr->wValue = le16_to_cpup(&dr->wValue);
	dr->wIndex = le16_to_cpup(&dr->wIndex);
	dr->wLength = le16_to_cpup(&dr->wLength);

#if 0
	{int i;
		for (i=0; i< UsbRecvLength; i++) {
			dbg("%s: pDataReceived[%d]=0x%02x", __FUNCTION__, i, pDataReceived[i]);
		}
	}
#endif

	if (pUsbOsHal->hUpper)
	{
		pUsbOsHal->cbUsbHalNotificationReceived (pUsbOsHal->hUpper, pDataReceived, UsbRecvLength);
	}
}

/****************************************************************
 *                                                              *
 *      NAME:                                                   *
 *      PARAMETERS: Standard for completion routine,			*
 *          Context points to the URB							*
 *                                                              *
 *      USAGE:                                                  *
 *                                                              *
 *      DESCRIPTION:                                            *
 *                                                              *
 *      NOTES:          None.                                   *
 ****************************************************************/
static void UsbTransmitDataCompletionRoutine(PURB urb)
{
	PURB_LIST_ENTRY ule = urb->context;
	PUSBOSHAL pUsbOsHal;
	int UsbStatus = urb->status;

	dbg("%s: ule=%p urb=%p UsbStatus=%d", __FUNCTION__, ule, urb, UsbStatus);

	if((UsbStatus == -ECONNRESET) || (UsbStatus == -ENOENT)) {
		dbg("%s: -ECONNRESET or -ENOENT", __FUNCTION__);
		ChangeUrbListEntryState(ule, ULE_ABORTED);
		return;
	}

	if(!ule)
		return;

	pUsbOsHal = ule->pUsbOsHal;
	CHECK_SIGNATURE (pUsbOsHal);

	if (UsbStatus)
	{
		err("%s: error: URB status %d", __FUNCTION__, UsbStatus);
	}

	if (pUsbOsHal->hUpper && !UsbStatus)
		pUsbOsHal->cbUsbHalDataTransmitDone (pUsbOsHal->hUpper);

	ChangeUrbListEntryState(ule, UsbStatus ? ULE_FAILED: ULE_COMPLETED);
}

/****************************************************************
 *                                                              *
 *      NAME:                                                   *
 *      PARAMETERS: Standard for completion routine,			*
 *          Context points to the URB							*
 *                                                              *
 *      USAGE:                                                  *
 *                                                              *
 *      DESCRIPTION:                                            *
 *                                                              *
 *      NOTES:          None.                                   *
 ****************************************************************/
static void UsbReceiveDataCompletionRoutine(PURB urb)
{
	PURB_LIST_ENTRY ule = urb->context;
	PUSBOSHAL pUsbOsHal;
	int UsbStatus = urb->status;
	// save a copy of data and received length
    int UsbRecvLength = urb->actual_length;
	PUINT8 pDataReceived = urb->transfer_buffer;

	dbg("%s: ule=%p urb=%p UsbStatus=%d UsbRecvLength=%d", __FUNCTION__, ule, urb, UsbStatus, UsbRecvLength);

	if((UsbStatus == -ECONNRESET) || (UsbStatus == -ENOENT)) {
		dbg("%s: -ECONNRESET or -ENOENT", __FUNCTION__);
		ChangeUrbListEntryState(ule, ULE_ABORTED);
		return;
	}

	if(!ule)
		return;

	pUsbOsHal = ule->pUsbOsHal;
	CHECK_SIGNATURE (pUsbOsHal);

	if (UsbStatus)
	{
		err("%s: error: URB status %d", __FUNCTION__, UsbStatus);
	}

	if (pUsbOsHal->hUpper && !UsbStatus) {
		if(ule->userBufferPtr) {
			memcpy(ule->userBufferPtr, pDataReceived, UsbRecvLength);
			pDataReceived = ule->userBufferPtr;
		}
		pUsbOsHal->cbUsbHalDataReceived (pUsbOsHal->hUpper, pDataReceived, UsbRecvLength);
	}

	ChangeUrbListEntryState(ule, UsbStatus ? ULE_FAILED: ULE_COMPLETED);
}

/****************************************************************
 *                                                              *
 *      NAME:                                                   *
 *      PARAMETERS: Standard for completion routine,			*
 *          Context points to the URB							*
 *                                                              *
 *      USAGE:                                                  *
 *                                                              *
 *      DESCRIPTION:                                            *
 *                                                              *
 *      NOTES:          None.                                   *
 ****************************************************************/
static void UsbDownloadCompletionRoutine (PURB urb)
{
	PURB_LIST_ENTRY ule = urb->context;
	PUSBOSHAL pUsbOsHal;
	int UsbStatus = urb->status;

	dbg("%s: ule=%p urb=%p UsbStatus=%d", __FUNCTION__, ule, urb, UsbStatus);

	if((UsbStatus == -ECONNRESET) || (UsbStatus == -ENOENT)) {
		dbg("%s: -ECONNRESET or -ENOENT", __FUNCTION__);
		ChangeUrbListEntryState(ule, ULE_ABORTED);
		return;
	}

	if(!ule)
		return;

	pUsbOsHal = ule->pUsbOsHal;
	CHECK_SIGNATURE (pUsbOsHal);

	if (UsbStatus)
	{
		err("%s: error: URB status %d", __FUNCTION__, UsbStatus);
	}

	if (pUsbOsHal->hUpper && !UsbStatus)
		pUsbOsHal->cbUsbHalDownloadTransmitDone (pUsbOsHal->hUpper);

	ChangeUrbListEntryState(ule, UsbStatus ? ULE_FAILED: ULE_COMPLETED);
}

/******************************************************************************

Name:           OsUsbMakeDataReceiveRequest

Description:    Schedules device-to-host bulk request on pipe 1 (data).

Arguments:      hUsbOsHal	- bottom HAL handle
				pBuf		- points to a buffer to store received data
				nBytes		- size of data to receive

Return Value:   result of the call

Notes:			This function can be called on high CPU priority level and must return
				immediately and must not wait on any synchronization object.

******************************************************************************/
BOOL OsUsbMakeDataReceiveRequest	(HANDLE hUsbOsHal, PVOID pBuf, UINT32 nBytes)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	PURB_LIST_ENTRY ule;

	dbg("%s: %p pBuf=%p nBytes=%ld", __FUNCTION__, hUsbOsHal, pBuf, nBytes);

	CHECK_SIGNATURE (pUsbOsHal);
	if (Working != GET_DEVICE_PNP_STATE(pUsbOsHal))
		return FALSE;

	// extract an URB from the list
	ule = GetFreeUrbFromList (&pUsbOsHal->RecvUrbList);
	if(!ule) {
		// normal?
		dbg("%s: no more free urbs", __FUNCTION__);
		return FALSE;
	}

	if(!IsKMallocMem(pBuf)) {
		ASSERT(nBytes <= ule->bufferSize);

		ule->userBufferPtr = pBuf;
		pBuf = ule->bufferPtr;
	} else
		ule->userBufferPtr = NULL;

	FILL_BULK_URB(ule->urb, pUsbOsHal->pUsbDevice, pUsbOsHal->DataInPipe,
			pBuf,
			nBytes,
			UsbReceiveDataCompletionRoutine,
			ule);


	return SubmitUrbListEntry(ule);
}

/******************************************************************************

Name:           OsUsbMakeDataTransmitRequest

Description:    Schedules host-to-device bulk request on pipe 1 (data).
				
Arguments:      hUsbOsHal	- bottom HAL handle
				hRequest	- handle to an OS specific data structure needed to 
							  initiate a bulk request. This handle is returned by
							  OsUsbGetTxRequest() call
				pBuf		- points to data to be transmitted
				nBytes		- size of data to be transmitted

Return Value:   result of the call

Notes:			This function can be called on high CPU priority level and must return
				immediately and must not wait on any synchronization object.

******************************************************************************/
BOOL OsUsbMakeDataTransmitRequest	(HANDLE hUsbOsHal, HANDLE hRequest, PVOID pBuf, UINT32 nBytes)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	PURB_LIST_ENTRY ule = (PURB_LIST_ENTRY)hRequest;

	dbg("%s: ule=%p pBuf=%p nBytes=%ld", __FUNCTION__, hRequest, pBuf, nBytes);

	CHECK_SIGNATURE (pUsbOsHal);
	if (Working != GET_DEVICE_PNP_STATE(pUsbOsHal))
		return FALSE;

	if(!IsKMallocMem(pBuf)) {
		ASSERT(nBytes <= ule->bufferSize);
		memcpy(ule->bufferPtr, pBuf, nBytes);
		pBuf = ule->bufferPtr;
	}

	FILL_BULK_URB(ule->urb, pUsbOsHal->pUsbDevice, pUsbOsHal->DataOutPipe,
			pBuf,
			nBytes,
			UsbTransmitDataCompletionRoutine,
			ule);

	return SubmitUrbListEntry(ule);
}

/******************************************************************************

Name:           OsUsbMakeDownloadTransmitRequest

Description:    Schedules host-to-device bulk request on pipe 3 (DSP download).
				
Arguments:      hUsbOsHal	- bottom HAL handle
				pBuf		- points to data to be transmitted
				nBytes		- size of data to be transmitted

Return Value:   result of the call

Notes:			This function can be called on high CPU priority level and must return
				immediately and must not wait on any synchronization object.

******************************************************************************/
BOOL OsUsbMakeDownloadTransmitRequest	(HANDLE hUsbOsHal, PVOID pBuf, UINT32 nBytes)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	PURB_LIST_ENTRY ule;

	dbg("%s: %p pBuf=%p nBytes=%ld", __FUNCTION__, hUsbOsHal, pBuf, nBytes);

	CHECK_SIGNATURE (pUsbOsHal);
	if (Working != GET_DEVICE_PNP_STATE(pUsbOsHal))
		return FALSE;

	// extract an URB from the list
	ule = GetFreeUrbFromList (&pUsbOsHal->DownloadUrbList);
	if(!ule) {
		err("%s: no more free urbs", __FUNCTION__);
		return FALSE;
	}

	if(!IsKMallocMem(pBuf)) {
		ASSERT(nBytes <= ule->bufferSize);
		memcpy(ule->bufferPtr, pBuf, nBytes);
		pBuf = ule->bufferPtr;
	}

	FILL_BULK_URB(ule->urb, pUsbOsHal->pUsbDevice, pUsbOsHal->DownloadPipe,
			pBuf,
			nBytes,
			UsbDownloadCompletionRoutine,
			ule);

//	ule->urb->transfer_flags |= USB_QUEUE_BULK; // XXX

	return SubmitUrbListEntry(ule);
}

/******************************************************************************

Name:           OsUsbWaitControlRequest

Description:    Waits until bottom HAL is ready to accept new control request.
				Returns a handle to a data structure needed to send such request.
				Wait must be accomplished by blocking a calling thread on some sort
				of syncronization object, like semaphore, so system is not locked and
				USB completions can be called by an OS

Arguments:      hUsbOsHal	- bottom HAL handle

Return Value:   hCntrlReq	- handle to an OS specific data structure needed to
							  initiate a control request.

Notes:

******************************************************************************/
HANDLE OsUsbWaitControlRequest (HANDLE hUsbOsHal)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	PURB_LIST_ENTRY ule;

	dbg("%s: %p", __FUNCTION__, hUsbOsHal);

	while(!(ule = GetFreeUrbFromList (&pUsbOsHal->ControlUrbList)) &&
			(Working == GET_DEVICE_PNP_STATE(pUsbOsHal))) {
		if( WAIT_TIMEOUT == OsWaitForEventTime((HOSEVENT)&pUsbOsHal->ControlRequestEvent, 5000)) { // give it 5s
			warn("OsWaitForControlRequestAvailable: timeout");
			BREAKPOINT ("OsWaitForControlRequestAvailable: timeout");
		}
	}

	dbg("%s: returning %p", __FUNCTION__, ule);
	return (HANDLE)ule;
}

/******************************************************************************

Name:           OsUsbWaitForControlRequestCompletion

Description:    Waits until control request is completed

Arguments:      hUsbOsHal	- bottom HAL handle

Return Value:   None.

Notes:

******************************************************************************/
void OsUsbWaitForControlRequestCompletion(HANDLE hUsbOsHal)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;

	dbg("%s: %p", __FUNCTION__, hUsbOsHal);

	if( WAIT_TIMEOUT == OsWaitForEventTime((HOSEVENT)&pUsbOsHal->ControlRequestEvent, 5000)) { // give it 5s
		err("OsUsbWaitForControlRequestCompletion: !! Control Request Timeout");
	}
}


/******************************************************************************

Name:           OsUsbGetControlRequest

Description:    Reserves a data structure needed to send a control request.

Arguments:      hUsbOsHal	- bottom HAL handle

Return Value:   handle to an OS specific data structure needed to
				initiate a control request or NULL if none available

Notes:			This function can be called on high CPU priority level and must return
				immediately and must not wait on any synchronization object.

******************************************************************************/
HANDLE OsUsbGetControlRequest (HANDLE hUsbOsHal)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	HANDLE h;

	CHECK_SIGNATURE (pUsbOsHal);
	h = (HANDLE)GetFreeUrbFromList (&pUsbOsHal->ControlUrbList);
	dbg("%s: %p returning %p", __FUNCTION__, hUsbOsHal, h);
	return h;
}

/******************************************************************************

Name:           OsUsbGetTxRequest

Description:    Reserves a data structure needed to send a host-to-device bulk request
				on pipe 1.

Arguments:      hUsbOsHal	- bottom HAL handle

Return Value:   handle to an OS specific data structure needed to
				initiate a bulk request or NULL if none available

Notes:			This function can be called on high CPU priority level and must return
				immediately and must not wait on any synchronization object.

******************************************************************************/
HANDLE OsUsbGetTxRequest (HANDLE hUsbOsHal)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	HANDLE h;

	CHECK_SIGNATURE (pUsbOsHal);
	h = (HANDLE)GetFreeUrbFromList (&pUsbOsHal->SendUrbList);
	dbg("%s: %p returning %p", __FUNCTION__, hUsbOsHal, h);
	return h;
}

/******************************************************************************

Name:           OsUsbMakeControlRequest

Description:    Formats a host-to-device control request and sends it via the default control 
				pipe (pipe 0). Does not wait for control request to be completed by USB

Arguments:      hUsbOsHal	- bottom HAL handle
				hCntrlReq	- handle to an OS specific data structure needed to
							  initiate a control request. Must be returned via previous
							  call to OsUsbWaitControlRequest() or OsUsbGetControlRequest()
				Request		- bRequest field of a control request header
				Value		- wValue field of a control request header
				Index		- wIndex field of a control request header
				pBuf		- pointer to data to be transmitted
				nBytes		- size of data pointed by pBuf

Return Value:   TRUE if request is being accepted by USB stack, FALSE if not

Notes:			Top HAL can call this function from a context where memory allocations
				are not permitted, such as DPC level. Since it is expected that bottom HAL
				has finite number of control structures (URBs/PBs) needed to initiate a USB
				transaction, top HAL first reserves such structure via OsUsbWaitControlRequest()
				or OsUsbGetControlRequest().

******************************************************************************/
BOOL OsUsbMakeControlRequest (HANDLE hUsbOsHal, HANDLE hCntrlReq, UINT8 Request, UINT16 Value, UINT16 Index,
							  IN PVOID pBuf, IN UINT32 nBytes)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	PURB_LIST_ENTRY ule;
	struct usb_ctrlrequest *dr;
	UINT16 snBytes = (UINT16)nBytes;

	dbg("%s: ule=%p req=0x%x val=%d idx=%d pbuf=%p len=%ld", __FUNCTION__, hCntrlReq, Request, Value, Index, pBuf, nBytes);

	CHECK_SIGNATURE (pUsbOsHal);

	if (Working != GET_DEVICE_PNP_STATE(pUsbOsHal))
		return FALSE;

#if 0
	if(nBytes > pUsbOsHal->pUsbDevice->descriptor.bMaxPacketSize0) {
//		warn("%s: nBytes (%ld) larger than MaxPacketSize (%d)", __FUNCTION__, nBytes, pUsbOsHal->pUsbDevice->descriptor.bMaxPacketSize0);
#if 0
		ASSERT(nBytes <= pUsbOsHal->pUsbDevice->descriptor.bMaxPacketSize0);
		nBytes = pUsbOsHal->pUsbDevice->descriptor.bMaxPacketSize0;
#endif
	}
#endif

	ule = (PURB_LIST_ENTRY)hCntrlReq;
	if (!ule) {
		err("%s: not enough control URBs?", __FUNCTION__);
		ASSERT(ule != NULL);
		return FALSE;
	}

	if(0 == ule->ul->count[ULE_FREE])
		OsClearEvent ((HOSEVENT)&pUsbOsHal->ControlRequestEvent);

	ASSERT(!(1 == Index && nBytes == 0));

	if(!IsKMallocMem(pBuf)) {
		ASSERT(nBytes <= ule->bufferSize);
		memcpy(ule->bufferPtr, pBuf, nBytes);
		pBuf = ule->bufferPtr;
	}

	dr = &ule->devRequest;
	dr->bRequestType = USB_TYPE_VENDOR | USB_RECIP_INTERFACE | USB_DIR_OUT;
	dr->bRequest = Request;

	dr->wValue = cpu_to_le16p(&Value);
	dr->wIndex = cpu_to_le16p(&Index);
	dr->wLength = cpu_to_le16p(&snBytes);

	pUsbOsHal->pLastControlData = pBuf;
	pUsbOsHal->nLastControlSize = nBytes;

	FILL_CONTROL_URB(ule->urb, pUsbOsHal->pUsbDevice,
			usb_sndctrlpipe(pUsbOsHal->pUsbDevice, 0),
			(unsigned char*)dr,
			pBuf,
			nBytes,
			UsbControlCompletionRoutine,
			ule);

	return SubmitUrbListEntry(ule);
}

/******************************************************************************

Name:           OsUsbMakeControlRequestSync

Description:    Formats a host-to-device control request and sends it via the default control 
				pipe (pipe 0). Wait for control request to be completed by USB

Arguments:      hUsbOsHal	- bottom HAL handle
				Direction	- control request direction: 0 - out, otherwise - in
				Request		- bRequest field of a control request header
				Value		- wValue field of a control request header
				Index		- wIndex field of a control request header
				pBuf		- pointer to data to be transmitted
				pnBytes		- pointer to size of data pointed by pBuf

Return Value:   TRUE if request is being accepted by USB stack, FALSE if not

Notes:			Top HAL can call this function from a context where memory allocations
				are not permitted, such as DPC level. Since it is expected that bottom HAL
				has finite number of control structures (URBs/PBs) needed to initiate a USB
				transaction, top HAL first reserves such structure via OsUsbWaitControlRequest()
				or OsUsbGetControlRequest().

******************************************************************************/
BOOL OsUsbMakeControlRequestSync		(HANDLE hUsbOsHal, IN UINT8 Direction, UINT8 Request, UINT16 Value, UINT16 Index,
										IN PVOID pBuf, IN OUT PUINT32 pnBytes)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	int ret;
	UINT8 ControlRequestBuffer[MAX_CONTROL_DATA_PACKET];

	dbg("%s: req=0x%x dir=%d val=%d idx=%d pbuf=%p len=%ld", __FUNCTION__, Request, Direction, Value, Index, pBuf, *pnBytes);

	if(*pnBytes > pUsbOsHal->pUsbDevice->descriptor.bMaxPacketSize0) {
		warn("%s: nBytes (%ld) larger than MaxPacketSize (%d)", __FUNCTION__, *pnBytes, pUsbOsHal->pUsbDevice->descriptor.bMaxPacketSize0);
		ASSERT(*pnBytes <= pUsbOsHal->pUsbDevice->descriptor.bMaxPacketSize0);
		*pnBytes = pUsbOsHal->pUsbDevice->descriptor.bMaxPacketSize0;
	}

	if(!IsKMallocMem(pBuf)) {
		ASSERT(*pnBytes <= sizeof(ControlRequestBuffer));
		memcpy(ControlRequestBuffer, pBuf, *pnBytes);
		pBuf = ControlRequestBuffer;
	}

	ret = usb_control_msg(pUsbOsHal->pUsbDevice,
			Direction ? usb_rcvctrlpipe(pUsbOsHal->pUsbDevice, 0) : usb_sndctrlpipe(pUsbOsHal->pUsbDevice, 0),
			Request,
			USB_TYPE_VENDOR | USB_RECIP_INTERFACE | (Direction ? USB_DIR_IN : USB_DIR_OUT),
			Value,
			Index,
			pBuf,
			*pnBytes,
			10*HZ);

	if(ret >= 0) {
		*pnBytes = ret;
		return TRUE;
	} else {
		err("%s: ERROR!!! req=0x%x err=%d", __FUNCTION__, Request, ret);
		return FALSE;
	}
}
//

#if 0
BOOL OsUsbMakeGetStringDescriptorRequestSync(
	HANDLE hUsbOsHal,
	IN UINT8 uStrIndex,	// String Descriptor Index
	IN PVOID pBuf,
	IN OUT PUINT32 pnBytes)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	int ret;

	ret = usb_string(pUsbOsHal->pUsbDevice, uStrIndex, pBuf, *pnBytes);
	if(ret < 0) {
		err"ERROR!!! OsUsbMakeGetStringDescriptorRequestSync: %d", ret);
		return FALSE;
	}

	*pnBytes = ret;
	return TRUE;
}
#endif
//
			
/******************************************************************************

Name:           OsUsbMakeNotifyRequest

Description:    Schedules device-to-host interrupt request on pipe 2 (notification).

Arguments:      hUsbOsHal	- bottom HAL handle

Return Value:   result of the call

Notes:			This function can be called on high CPU priority level and must return
				immediately and must not wait on any synchronization object.

******************************************************************************/
BOOL OsUsbMakeNotifyRequest	(HANDLE hUsbOsHal)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;
	PURB_LIST_ENTRY ule;

	dbg("%s: %p", __FUNCTION__, hUsbOsHal);

	CHECK_SIGNATURE (pUsbOsHal);
	if (Working != GET_DEVICE_PNP_STATE(pUsbOsHal))
		return FALSE;

	// extract an URB from the list
	ule = GetFreeUrbFromList (&pUsbOsHal->NotifyUrbList);
	if(!ule) {
		// normal?
		dbg("%s: no more free urbs; request already active?", __FUNCTION__);
		return FALSE;
	}

	FILL_INT_URB(ule->urb, pUsbOsHal->pUsbDevice, pUsbOsHal->NotifyPipe,
			ule->bufferPtr,
			ule->bufferSize,
			UsbNotifyCompletionRoutine,
			ule,
			pUsbOsHal->NotifyInterval);

	ule->urb->transfer_flags |= USB_ASYNC_UNLINK;

	return SubmitUrbListEntry(ule);
}

/******************************************************************************

Name:           OsUsbResetPipes

Description:    Aborts any transactions-in-progress across USB

Arguments:      hUsbOsHal	- bottom HAL handle

Return Value:   none

Notes:

******************************************************************************/
void OsUsbResetPipes (HANDLE hUsbOsHal)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hUsbOsHal;

	dbg("%s: %p", __FUNCTION__, hUsbOsHal);
//	pUsbOsHal->bPipeResetInProgress = TRUE;

	// XXX if (NULL == pUsbOsHal->pDownloadUrb || NULL == pUsbOsHal->pDownloadUrb->next)
	if(0 == pUsbOsHal->DownloadUrbList.count[ULE_FREE]) {
		err("Download pipe stuck, resetting");
        usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DownloadPipe);
	}

//	pUsbOsHal->bPipeResetInProgress = FALSE;
	// To Do; if we have some stuck pipes here, reset them
}

static void OsUsbFreeUrbs (PUSBOSHAL pUsbOsHal)
{
	dbg("%s: %p", __FUNCTION__, pUsbOsHal);

	FreeUrbList(&pUsbOsHal->NotifyUrbList);
	FreeUrbList(&pUsbOsHal->RecvUrbList);
	FreeUrbList(&pUsbOsHal->SendUrbList);
	FreeUrbList(&pUsbOsHal->ControlUrbList);
	FreeUrbList(&pUsbOsHal->DownloadUrbList);
#ifdef USE_DCP
	FreeUrbList(&pUsbOsHal->DcpUrbList);
#endif
}

static BOOL OsUsbAllocateUrbs (PUSBOSHAL pUsbOsHal)
{
	dbg("%s: %p", __FUNCTION__, pUsbOsHal);

	// for now allocate just one IRP/URB per pipe

	if(!InitUrbList(&pUsbOsHal->RecvUrbList, pUsbOsHal,
				MAX_RECEIVE_URB_BUFSIZE, NUM_RECEIVE_URBS)) {
		goto error;
	}

	if(!InitUrbList(&pUsbOsHal->SendUrbList, pUsbOsHal,
				MAX_SEND_URB_BUFSIZE, NUM_SEND_URBS)) {
		goto error;
	}

	if(!InitUrbList(&pUsbOsHal->NotifyUrbList, pUsbOsHal,
				MAX_NOTIFY_URB_BUFSIZE, NUM_NOTIFY_URBS)) {
		goto error;
	}

	if(!InitUrbList(&pUsbOsHal->ControlUrbList, pUsbOsHal,
				MAX_CONTROL_URB_BUFSIZE, NUM_CONTROL_URBS)) {
		goto error;
	}

	if(!InitUrbList(&pUsbOsHal->DownloadUrbList, pUsbOsHal,
				MAX_DOWNLOAD_URB_BUFSIZE, NUM_DOWNLOAD_URBS)) {
		goto error;
	}

#ifdef USE_DCP
	if(!InitUrbList(&pUsbOsHal->DcpUrbList, pUsbOsHal,
				MAX_DCP_URB_BUFSIZE, NUM_DCP_URBS)) {
		goto error;
	}
#endif

	return TRUE;

error:
	OsUsbFreeUrbs(pUsbOsHal);
	return FALSE;
}

#if 0
static void ReturnUrbToList(volatile PURB * ppUrb, PURB urb)
{
    PURB oldhead;

	dbg("%s: %p to %p", __FUNCTION__, urb, ppUrb);
    do 
	{
        oldhead = *ppUrb;
        urb->next = oldhead;
    } while(!OsAtomicCompareAndSwap(oldhead, urb, (PVOID *)ppUrb));
}

static PURB GetUrbFromList(volatile PURB * ppUrb)
{
    PURB urb, urb1;

    do {
        urb = * ppUrb;
        if (NULL == urb)
            {
            return NULL;
            }
        urb1 = urb->next;
        } 
    while(!OsAtomicCompareAndSwap(urb, urb1, (PVOID *)ppUrb));
    urb->next = NULL;

	dbg("%s: got %p from %p", __FUNCTION__, urb, ppUrb);
    return urb;
}
#endif

/******************************************************************************

Name:           OsInitEvent

Description:    Initializes an OS-specific syncronization object with event capability

Arguments:      none

Return Value:   handle to event

Notes:

******************************************************************************/
void		OsInitEvent	(PEVENT pEvent)
{
	init_waitqueue_head(&pEvent->wq);
#if 0 // XXX
	pEvent->flag = 1; // Create signalled
#else
	pEvent->flag = 0;
#endif
	spin_lock_init(&pEvent->lock);
}

/******************************************************************************

Name:           OsCreateEvent

Description:    Creates an OS-specific syncronization object with event capability

Arguments:      none

Return Value:   handle to created event

Notes:

******************************************************************************/
HOSEVENT	OsCreateEvent	(void)
{
	PEVENT pEvent = (PEVENT) kmalloc(sizeof(EVENT), GFP_KERNEL);

	if(pEvent)
		OsInitEvent(pEvent);

	dbg("%s: %p", __FUNCTION__, pEvent);
	return (HOSEVENT)pEvent;
}

/******************************************************************************

Name:           OsDeleteEvent

Description:    Destroys event object

Arguments:      hEvent	- handle to event

Return Value:   none

Notes:

******************************************************************************/
void		OsDeleteEvent	(HOSEVENT hEvent)
{
	PEVENT pEvent = (PEVENT)hEvent;

	dbg("%s: %p", __FUNCTION__, hEvent);
	kfree(pEvent);
}

/******************************************************************************

Name:           OsWaitForEvent

Description:    waits until event is set by OsSetEvent() call
				
Arguments:      hEvent	- handle to event

Return Value:   none

Notes:

******************************************************************************/
void		OsWaitForEvent	(HOSEVENT hEvent)
{
#if 1 // XXX
	dbg("%s: %p", __FUNCTION__, hEvent);
	OsWaitForEventTime(hEvent, 10*1000);
#else
	PEVENT pEvent = (PEVENT)hEvent;
	DECLARE_WAITQUEUE(wait, current);
	unsigned long flags;

	dbg("%s: %p", __FUNCTION__, hEvent);

	add_wait_queue(&pEvent->wq, &wait);
	while(1) {
		spin_lock_irqsave(&pEvent->lock, flags);
		if(pEvent->flag) {
			pEvent->flag = 0;
			spin_unlock_irqrestore(&pEvent->lock, flags);
			break;
		}
		spin_unlock_irqrestore(&pEvent->lock, flags);
		rmb();
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&pEvent->wq, &wait);
#endif
}

/******************************************************************************

Name:           OsSetEvent

Description:    sets event to signalled state

Arguments:      hEvent	- handle to event

Return Value:   none

Notes:

******************************************************************************/
void		OsSetEvent		(HOSEVENT hEvent)
{
	PEVENT pEvent = (PEVENT)hEvent;
	unsigned long flags;

	dbg("%s: %p", __FUNCTION__, hEvent);
	spin_lock_irqsave(&pEvent->lock, flags);
	pEvent->flag = 1;
	spin_unlock_irqrestore(&pEvent->lock, flags);
	wmb();
	wake_up(&pEvent->wq);
}

/***************************************************************************/

static inline long MSECS_TO_TICKS(UINT32 msecs)
{
	long ticks = (msecs)/(1000/HZ);

	// round up to next tick
	if ((!ticks && msecs) || (msecs % (1000/HZ)) >= ((1000/HZ) / 2))
		ticks++;

	return ticks;
}

/****************************************************************************

Name:           OsWaitForEventTime

Description:    waits until event is set by OsSetEvent() call or specified time has passed

Arguments:      hEvent	- handle to event
				timeout	- time in ms for function to block the calling thread if event remains
				in non-signalled state

Return Value:   Result of the call
				WAIT_OK			- event gets set
				WAIT_TIMEOUT	- time elapses
				WAIT_ERROR		- error

Notes:

****************************************************************************/
WAIT_RESULT	OsWaitForEventTime	(HOSEVENT hEvent, UINT32 timeout)
{
	PEVENT pEvent = (PEVENT)hEvent;
	DECLARE_WAITQUEUE(wait, current);
	unsigned long flags;
	long tmticks = MSECS_TO_TICKS(timeout);
	WAIT_RESULT res = WAIT_TIMEOUT;

	dbg("%s: %p timeout=%ld", __FUNCTION__, hEvent, timeout);

	add_wait_queue(&pEvent->wq, &wait);
	do {
		spin_lock_irqsave(&pEvent->lock, flags);
		if(pEvent->flag) {
			pEvent->flag = 0;
			spin_unlock_irqrestore(&pEvent->lock, flags);
			res = WAIT_OK;
			break;
		}
		spin_unlock_irqrestore(&pEvent->lock, flags);
		rmb();
		set_current_state(TASK_UNINTERRUPTIBLE);
	} while ( tmticks && (tmticks = schedule_timeout(tmticks)) );

	current->state = TASK_RUNNING;
	remove_wait_queue(&pEvent->wq, &wait);

	if(res == WAIT_TIMEOUT) {
		dbg("%s: %p returning WAIT_TIMEOUT", __FUNCTION__, hEvent);
	}
	return res;
}

/******************************************************************************

Name:           OsClearEvent

Description:    sets event to non-signalled state
				
Arguments:      hEvent	- handle to event

Return Value:   none

Notes:

******************************************************************************/
void		OsClearEvent	(HOSEVENT hEvent)
{
	PEVENT pEvent = (PEVENT)hEvent;
	unsigned long flags;

	dbg("%s: %p", __FUNCTION__, hEvent);

	spin_lock_irqsave(&pEvent->lock, flags);
	pEvent->flag = 0;
	spin_unlock_irqrestore(&pEvent->lock, flags);
	wmb();
}


#ifdef USE_DCP

#error not yet supported under Linux

#define DCP_BLOCK_SIZE (1024*1024)

typedef struct tagDCP
{
    HANDLE	hHalOs;

    BOOL	fOn;
	int		nBufNo;

	PIO_WORKITEM	pWorkItem;

    UINT16  Buffer[2][DCP_BLOCK_SIZE];
    UINT32  dwCount;
	UINT32	hFile;

} DCP, *PDCP;


static DCP gDcp;

static void UsbDcpReceived ( int UsbRecvLength);
static NTSTATUS UsbDcpCompletionRoutine(PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context);


HANDLE DcpCreate(HANDLE hHalOs, HANDLE hDevNode)
{
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)hHalOs;
    gDcp.fOn        = FALSE;
    gDcp.dwCount    = 0;
    gDcp.hHalOs     = hHalOs;
	gDcp.nBufNo		= 0;

	gDcp.pWorkItem = IoAllocateWorkItem (pUsbOsHal->pDeviceExtension->DeviceObject);
    return &gDcp;
}

void DcpDestroy  (HANDLE hDcp)
{
    PDCP pDcp = (PDCP)hDcp;
	IoFreeWorkItem (pDcp->pWorkItem);
}

void DcpWorkItemRtn (PDEVICE_OBJECT pDevObject, PVOID Context)
{
    PDCP pDcp = (PDCP)Context;
	if (pDcp->fOn && pDcp->hFile)
		R0FileWrite(pDcp->hFile, (UINT8*)pDcp->Buffer[1-pDcp->nBufNo], sizeof(pDcp->Buffer[0]));
}

void DcpSetVolume(HANDLE hDcp, int nVolume)
{
    PDCP pDcp = (PDCP)hDcp;

    if( nVolume && !pDcp->fOn)
    {
		static unsigned fileNo;
        static char fileName[255];
		OsSprintf (fileName, "\\??\\C:\\temp\\dcp%d.snd", fileNo++);
        pDcp->hFile = R0FileOpen(fileName);

        pDcp->fOn = TRUE;
        pDcp->dwCount = 0;
        OsUsbMakeDcpRequest(pDcp->hHalOs, &pDcp->Buffer[pDcp->dwCount], 2048);
    }
    else if( !nVolume && pDcp->fOn )
    {
		pDcp->fOn = FALSE;
		if(pDcp->hFile)
		{
		   R0FileWrite(pDcp->hFile, (UINT8*)pDcp->Buffer[pDcp->nBufNo], 2*pDcp->dwCount);
		   R0FileClose(pDcp->hFile);
		   pDcp->hFile = 0;
		}
		// Flush data into file
		RTRACE1(T__ERROR, 0,"************* Recorded %d DCP samples", pDcp->dwCount );
    }
}

static void UsbDcpReceived ( int UsbRecvLength)
{
    PDCP pDcp = &gDcp;

    if( pDcp->fOn )
    {
        pDcp->dwCount += UsbRecvLength/2;
        
        if (pDcp->dwCount == DCP_BLOCK_SIZE)
		{	
			pDcp->nBufNo = 1-pDcp->nBufNo;
			IoQueueWorkItem (pDcp->pWorkItem, DcpWorkItemRtn, DelayedWorkQueue, pDcp);
			pDcp->dwCount = 0;
            OsUsbMakeDcpRequest(pDcp->hHalOs, pDcp->Buffer[pDcp->nBufNo], 2048);
		}
		else
            OsUsbMakeDcpRequest(pDcp->hHalOs, &pDcp->Buffer[pDcp->nBufNo][pDcp->dwCount], 2048);
			
    }
}

static NTSTATUS UsbDcpCompletionRoutine(
    IN PDEVICE_OBJECT DeviceObject,
       IN PIRP Irp,
       IN PVOID Context)
{
    PURB urb = (PURB) Context;
	PUSBOSHAL pUsbOsHal = (PUSBOSHAL)((PDEVICE_EXTENSION)DeviceObject->DeviceExtension)->HALExtension;

	// save a copy of data and received length
    int UsbRecvLength = urb->actual_length;

	CHECK_SIGNATURE (pUsbOsHal);
	if (UsbStatus)
	{
		RTRACE2(T__HALOS|T__ERROR, 0, "UsbDcpCompletionRoutine error: 0x%x, URB status 0x%x",
				Irp->IoStatus.Status, urb->UrbHeader.Status);
	}

    ReturnUrbToList (&pUsbOsHal->pDcpUrb, urb);
    ReturnIrpToList (&pUsbOsHal->pDcpIrp, Irp);

	if (!UsbStatus)
		UsbDcpReceived (UsbRecvLength);

	RTRACE0(T__HALOS, 0,"Exit UsbDcpCompletionRoutine");
}


/******************************************************************************

Name:           OsUsbMakeDcpRequest

Description:    Schedules device-to-host bulk request on pipe 3 (DCP).

Arguments:      hUsbOsHal	- bottom HAL handle
				pBuf		- buffer to read samples to
				nBytes		- size of such buffer

Return Value:   none

Notes:			After USB completion routine is called, bottom HAL reschedules USB
				bulk request on pipe 3 without top HAL intervention until
				OsUsbCancelDcpRequests() is called.

******************************************************************************/
BOOL OsUsbMakeDcpRequest(HANDLE hOsHal, PVOID pBuf, UINT32 nBytes)
{
	PIRP irp;
	PURB urb;
    PUSBOSHAL pUsbOsHal = hOsHal;

	CHECK_SIGNATURE (pUsbOsHal);
	// extract an URB from the list
    urb = GetUrbFromList (& pUsbOsHal->pDcpUrb);
	if (NULL == urb)
		return FALSE;

	// extract an IRP from the list
	irp = GetIrpFromList(&pUsbOsHal->pDcpIrp);
	if (NULL == irp)
	{
		// no IRP available or IRP is cancelled and deallocated
		RTRACE0(T__ERROR | T__HALOS, 0,"NULL == DcpIrp");
		ReturnUrbToList( & pUsbOsHal->pDcpUrb, urb);
		return FALSE;
	}

	if (STATUS_PENDING != CallUsbd (
		pUsbOsHal, irp, urb, pUsbOsHal->DcpInPipe,
		pBuf, nBytes, UsbDcpCompletionRoutine, CALL_USBD_FLAGS_INPUT))
		return FALSE;

    return TRUE;
}


#endif // USE_DCP


static void * usbhcf_probe(struct usb_device *pUsbDevice, unsigned int ifnum,
			    const struct usb_device_id *id)
{
	int i;
	struct usb_interface *pUsbDataInterface, *pUsbCommInterface;
	PUSBOSHAL pUsbOsHal;

	dbg("%s: pUsbDevice=%p ifnum=%d id=%p bConfigurationValue=%d", __FUNCTION__, pUsbDevice, ifnum, id, pUsbDevice->config[0].bConfigurationValue);

#if 0 // bad, only useful for fw dl testing
	// undo the usb_set_configuration(1) done automatically by usb_new_device()
	// since the device reacts weirdly (DCD & DATA LEDs come on)
	if(pUsbDevice->actconfig->bConfigurationValue != 0) {
		pUsbDevice->actconfig->bConfigurationValue = 0;
dbg("usbhcf_probe: REDOING usb_set_configuration()");
		if (usb_set_configuration(pUsbDevice, pUsbDevice->actconfig->bConfigurationValue)) {
			err("usb_set_configuration() failed");
			return NULL;
		}
	}
#endif

	if (((pUsbDevice->descriptor.idVendor != USB_AAPL_VENDOR_ID) &&
				(pUsbDevice->descriptor.idVendor != USB_CNXT_VENDOR_ID)) ||
	    ((pUsbDevice->descriptor.idProduct != USB_AAPL_PRODUCT_ID) &&
				 (pUsbDevice->descriptor.idProduct != USB_CNXT_PRODUCT_ID))) {
	    	err("Not the one we are interested about");
		return NULL;
	}

	if(pUsbDevice->descriptor.bNumConfigurations != 1) {
		err("Wrong number of device configurations (%d)", pUsbDevice->descriptor.bNumConfigurations);
		return NULL;
	}

	if(pUsbDevice->actconfig->bNumInterfaces != 2) {
		err("Wrong number of device interfaces (%d)", pUsbDevice->actconfig->bNumInterfaces);
		return NULL;
	}

	pUsbDataInterface = pUsbDevice->actconfig->interface + 0;
	if(pUsbDataInterface->num_altsetting != 1) {
		warn("DataInterface has more than one alternate setting (%d)", pUsbDataInterface->num_altsetting);
	}
	if(pUsbDataInterface->altsetting->bNumEndpoints != DATA_PIPE_NUM) {
		err("Wrong number of endpoints (%d) for DataInterface", pUsbDataInterface->altsetting->bNumEndpoints);
		return NULL;
	}

	pUsbCommInterface = pUsbDevice->actconfig->interface + 1;
	if(pUsbCommInterface->num_altsetting != 1) {
		warn("CommInterface has more than one alternate setting (%d)", pUsbCommInterface->num_altsetting);
	}
	if(pUsbCommInterface->altsetting->bNumEndpoints != INT_PIPE_NUM) {
		err("Wrong number of endpoints (%d) for CommInterface", pUsbCommInterface->altsetting->bNumEndpoints);
		return NULL;
	}

	pUsbOsHal = kmalloc(sizeof(USBOSHAL), GFP_KERNEL);
	if (!pUsbOsHal) {
		err ("Out of memory");
		goto exit;
	}
	memset(pUsbOsHal, 0, sizeof(USBOSHAL));

	for(i=0; i < pUsbDataInterface->altsetting->bNumEndpoints; i++) {
		dbg("%s: pUsbDataInterface endpoint=%d len=%d type=0x%02x addr=0x%02x attr=0x%02x maxsize=%d interval=%d",
			__FUNCTION__,
			i,
			pUsbDataInterface->altsetting->endpoint[i].bLength,
			pUsbDataInterface->altsetting->endpoint[i].bDescriptorType,
			pUsbDataInterface->altsetting->endpoint[i].bEndpointAddress,
			pUsbDataInterface->altsetting->endpoint[i].bmAttributes,
			pUsbDataInterface->altsetting->endpoint[i].wMaxPacketSize,
			pUsbDataInterface->altsetting->endpoint[i].bInterval
		);

		if (pUsbDataInterface->altsetting->endpoint[i].bEndpointAddress == 0x81) {
			pUsbOsHal->DataInPipe = usb_rcvbulkpipe (pUsbDevice, pUsbDataInterface->altsetting->endpoint[i].bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
		} else if (pUsbDataInterface->altsetting->endpoint[i].bEndpointAddress == 0x01) {
			pUsbOsHal->DataOutPipe = usb_sndbulkpipe (pUsbDevice, pUsbDataInterface->altsetting->endpoint[i].bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
		} else if (pUsbDataInterface->altsetting->endpoint[i].bEndpointAddress == 0x83) {
			pUsbOsHal->DcpInPipe = usb_rcvbulkpipe (pUsbDevice, pUsbDataInterface->altsetting->endpoint[i].bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
		} else if (pUsbDataInterface->altsetting->endpoint[i].bEndpointAddress == 0x03) {
			pUsbOsHal->DownloadPipe = usb_sndbulkpipe (pUsbDevice, pUsbDataInterface->altsetting->endpoint[i].bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
			if(pUsbDataInterface->altsetting->endpoint[i].wMaxPacketSize == 64)
				pUsbOsHal->UpdateEEPROM = TRUE;
		}
	}

	for(i=0; i < pUsbCommInterface->altsetting->bNumEndpoints; i++) {
		dbg("%s: pUsbCommInterface endpoint=%d len=%d type=0x%02x addr=0x%02x attr=0x%02x maxsize=%d interval=%d",
			__FUNCTION__,
			i,
			pUsbCommInterface->altsetting->endpoint[i].bLength,
			pUsbCommInterface->altsetting->endpoint[i].bDescriptorType,
			pUsbCommInterface->altsetting->endpoint[i].bEndpointAddress,
			pUsbCommInterface->altsetting->endpoint[i].bmAttributes,
			pUsbCommInterface->altsetting->endpoint[i].wMaxPacketSize,
			pUsbCommInterface->altsetting->endpoint[i].bInterval
		);

		if (pUsbCommInterface->altsetting->endpoint[i].bEndpointAddress == 0x82) {
			pUsbOsHal->NotifyPipe = usb_rcvintpipe (pUsbDevice, pUsbCommInterface->altsetting->endpoint[i].bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
			pUsbOsHal->NotifyInterval = pUsbCommInterface->altsetting->endpoint[i].bInterval;
			pUsbOsHal->MaxNotifySize = pUsbCommInterface->altsetting->endpoint[i].wMaxPacketSize;
		}
	}

	if(!pUsbOsHal->DataInPipe ||
			!pUsbOsHal->DataOutPipe ||
			!pUsbOsHal->DcpInPipe ||
			!pUsbOsHal->DownloadPipe ||
			!pUsbOsHal->NotifyPipe) {
		err("Missing endpoint(s)");
		kfree(pUsbOsHal);
		pUsbOsHal = NULL;
		goto exit;
	}

	if(!pUsbOsHal->NotifyInterval)
		pUsbOsHal->NotifyInterval = 1; // ms

	pUsbOsHal->pUsbDevice = pUsbDevice;
	pUsbOsHal->pUsbCommInterface = pUsbCommInterface;
	pUsbOsHal->pUsbDataInterface = pUsbDataInterface;

	OsInitEvent(&pUsbOsHal->ControlRequestEvent);

	if(!OsUsbAllocateUrbs(pUsbOsHal)) {
		err("Cannot allocate URBs");
		kfree(pUsbOsHal);
		pUsbOsHal = NULL;
		goto exit;
	}

	if(OsUSBFWDownload (pUsbOsHal)) {
		err("Firmware download failed");
	} else
		usb_started = TRUE;


	usb_driver_claim_interface(&usbhcf_driver, pUsbDevice->actconfig->interface + 1, pUsbOsHal);

	info("usbhcf is detected");

	list_add(&pUsbOsHal->instance_list_entry, &instance_list);

#if 0
	{
		ULONG nBytes = 0;
		static unsigned char scbuf[] = { 
			0x0B, 0x22, 0x20, 0x17, 0x21, 0x08, 0x17, 0x01, 0x88, 0x3D,
			0xFE, 0x64, 0x13, 0x21, 0x08, 0x13, 0x01, 0x88, 0x3D, 0xFE,
			0x32, 0x17, 0x01, 0x08, 0x3D, 0xFE, 0xC8, 0x3D, 0xFE, 0xC8,
			0x26, 0x36 };

		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DataInPipe);
		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DataOutPipe);
		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DcpInPipe);
		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->DownloadPipe);
		usb_clear_halt(pUsbOsHal->pUsbDevice, pUsbOsHal->NotifyPipe);

		OsUsbMakeNotifyRequest((HANDLE)pUsbOsHal);

		if (!OsUsbMakeControlRequestSync((HANDLE)pUsbOsHal, 0,
					CD2_RESET, 0, 0, NULL, &nBytes)) {
			err("CD2_RESET error nBytes=%ld", nBytes);
		}
		nBytes = sizeof(scbuf);
		if (!OsUsbMakeControlRequestSync((HANDLE)pUsbOsHal, 0,
					CD2_CONTROL_SCRIPT, 0, 1, scbuf, &nBytes)) {
			err("CD2_CONTROL_SCRIPT error nBytes=%ld", nBytes);
		}
	}
#endif
exit:
	return pUsbOsHal;
}

static void usbhcf_disconnect(struct usb_device *pUsbDevice, void *ptr)
{
	PUSBOSHAL pUsbOsHal = ptr;

	dbg("%s: %p", __FUNCTION__, pUsbOsHal);

	if(!pUsbOsHal)
		return;

	usb_driver_release_interface(&usbhcf_driver, pUsbDevice->actconfig->interface + 1);

	OsUsbFreeUrbs(pUsbOsHal);

	kfree(pUsbOsHal);
	pUsbOsHal = NULL;
}

#ifdef CONFIG_ALL_PPC
static int pmacftrmodemenable = 1;
MODULE_PARM(pmacftrmodemenable, "i");
MODULE_PARM_DESC(pmacftrmodemenable, "Enable internal modem if non-zero");
#endif

static int __init usbhcf_init(void)
{
	int ret;
	dbg("%s", __FUNCTION__);
#ifdef CONFIG_ALL_PPC
	if(pmacftrmodemenable) {
		ret = pmac_call_feature(PMAC_FTR_MODEM_ENABLE, NULL, 0, 1);
		if(ret) {
			err("%s: cannot enable internal modem (%d)", __FUNCTION__, ret);
			return ret;
		}
	}
#endif
	ret = usb_register(&usbhcf_driver);
	if(ret) {
		err("%s: usb_register (%d) failed", __FUNCTION__, ret);
	}
	return ret;
}


static void __exit usbhcf_exit(void)
{
	dbg("%s", __FUNCTION__);
	usb_deregister(&usbhcf_driver);
#ifdef CONFIG_ALL_PPC
	if(pmacftrmodemenable) {
		int ret = pmac_call_feature(PMAC_FTR_MODEM_ENABLE, NULL, 0, 0);
		if(ret) {
			err("%s: cannot disable internal modem (%d)", __FUNCTION__, ret);
			return;
		}
	}
#endif
	usb_started = FALSE;
}

module_init(usbhcf_init);
module_exit(usbhcf_exit);

